//---------------------------------------------------------------------------
// Copyright (C) 1999 Dallas Semiconductor Corporation, All Rights Reserved.
// 
// Permission is hereby granted, free of charge, to any person obtaining a 
// copy of this software and associated documentation files (the "Software"), 
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit persons to whom the 
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included 
// in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
// IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES 
// OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
// OTHER DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of Dallas Semiconductor 
// shall not be used except as stated in the Dallas Semiconductor 
// Branding Policy. 
//---------------------------------------------------------------------------
//
//  1wsnsut.C - 1-Wire Sensor utility functions.  Run scripts to perform
//              1-Wire sensor operations.  Use variables in MIB as needed.
//
//  Compiler:  Microsoft Visual C 
//  Version:   1.00
//  History:   
//

#include "ibtmex.h"
#include "1wsensor.h"   
#include "objects.h"
#include "tagconst.h"

// external function prototypes
extern int ReadFromMIB(ushort, uchar *, int);
extern void output_status(int, char *);

// Local Function Prototypes 
int RunScript(Script script[], void *, int);
int ScriptStep(Script *, int *, int *, int *, char *, void *, int);
int DeviceSetVerify(void);
int Read1820Temp(float *);
int Set2406State(void);
int Read2423Counter(ulong *);
int ReadTagNext(int *, uchar *, int);
uchar dowcrc(uchar);
ushort docrc16(ushort);
int ROMToStr(char *, uchar *);
int ROMSwitchState(uchar *);
int bitacc(int, int, int, uchar *);
int DeviceCheckedForTag(uchar **);
int Read2406Info(int, uchar *);
int Send2409Command(int, int);
int Set2409State(void);
int Read2409Info(uchar *);

// defines
#define ROM_STACK_MAX   200
#define RETRY_MAX       5

// globals
uchar DOWCRC;
static short rom_stack[ROM_STACK_MAX][8];
static int stack_num=0;
static uchar retry_stack[ROM_STACK_MAX];

//--------------------------------------------------------------------------
// Script functions
//--------------------------------------------------------------------------

//--------------------------------------------------------------------------
// Run the specified script. 
// Status is printed using 'output_status' function.  The script result is
// returned in the buffer 'prslt' of max length 'rslt_max'.  Result data 
// format is script dependent.
//
// Return: TRUE - all steps completed
//         FALSE - all steps NOT completed due to error or escape 
//
int RunScript(Script script[], void *prslt, int rslt_max)
{
   char statusmsg[256],LastDescription[256],LastMsg[256];
   char tmsg[256];
   int StepCount,SubStep,ErrorCount,Status,i;
   uchar *puchar;   

   // reset the step to the begining
   StepCount = 0;
   SubStep = 0;
   ErrorCount = 0;
   Status = STATUS_INPROGRESS;
   statusmsg[0] = 0;
   puchar = prslt;
   for (i = 0; i < rslt_max; i++)
      puchar[i] = 0;
 
   // loop to perform all of the steps to download the Thermochron
   do
   {   
      // switch on the status of the last step done
      switch(Status)
      {
         // step complete so go to the next
         case STATUS_STEP_COMPLETE:
            StepCount++;
            SubStep = 0;
            ErrorCount = 0;
            Status = STATUS_INPROGRESS;
            LastDescription[0] = 0;
            LastMsg[0] = 0;
            break;
         // in progress so call again
         case STATUS_INPROGRESS:
            // print step description if different 
            if (strcmp(LastDescription,
                script[StepCount].StepDescription) != 0)
            {
               sprintf(tmsg,"%s --> ",script[StepCount].StepDescription);
               output_status(LV_VERBOSE,tmsg);
               sprintf(LastDescription,"%s",script[StepCount].StepDescription);
            }

            // perform a step in the job
            ScriptStep(&script[StepCount],&SubStep, 
                      &Status, &ErrorCount, statusmsg, 
                      prslt, rslt_max);

            // print results if different
            if (strcmp(LastMsg,statusmsg) != 0)
            {
               sprintf(tmsg,"%s\n",statusmsg);
               output_status(LV_VERBOSE,tmsg);
               sprintf(LastMsg,"%s",statusmsg);
            }
            else
               output_status(LV_VERBOSE,".");
            break;
         // encountered a transient error
         case STATUS_ERROR_TRANSIENT:
            // if 20 tansient errors in a row then abort
            if (ErrorCount > 20)
               Status = STATUS_ERROR_HALT;
            else
               Status = STATUS_INPROGRESS;
            break;
         // all steps complete
         case STATUS_COMPLETE:
            output_status(LV_VERBOSE,"End script normally\n");
            return TRUE;
         // non-recoverable error
         case STATUS_ERROR_HALT:
            // try to do emergency clean up
            TMEndSession(session_handle);
            output_status(LV_ALWAYS,"Aborting script due to non-recoverable error\n");
            return FALSE;
      }
   }
   while (!kbhit());

   // key abort
   output_status(LV_ALWAYS,"Aborting script due to key press\n");
   return FALSE;
}

//----------------------------------------------------------------------
//  Use the script to perform a step and return.
//
//  Return:
//          STATUS_STEP_COMPLETE - step complete, go to next step
//          STATUS_INPROGRESS - step is not complete, call again
//          STATUS_ERROR_HALT -  error, halt script
//          STATUS_ERROR_TRANSIENT - transient error, call again
//
int ScriptStep(Script *StateScript, int *SubStep,
              int *Status, int *ErrorCount, char *statusmsg,
              void *prslt, int rslt_max)
{
   short rslt;
   int *pint;
   float tmp, *pfloat;
   uchar tbuf[255],*puchar;
   int scnt,obln,state,i;   
   ulong tulong, *pulong;
   short trom[8];

   // do the current step and return
   switch (StateScript->Step)
   {
      // attempt to get a session on the 1-Wire Net
      case ST_GET_SESSION:
         session_handle = TMExtendedStartSession(PortNum,PortType,NULL);
         if (session_handle > 0)
         {
            sprintf(statusmsg,"Session aquired");
            *Status = STATUS_STEP_COMPLETE;
         }
         else if (session_handle == 0)
         {
            sprintf(statusmsg,"1-Wire Net in use by other process");
            *Status = STATUS_INPROGRESS;
         }
         else if (session_handle < 0)
         {
            sprintf(statusmsg,"1-Wire Net not available");
            *Status = STATUS_ERROR_HALT;
         }
         break;
      
      // perform TMSetup on the 1-Wire Net port   
      case ST_SETUP:
         rslt = TMSetup(session_handle);
         if (rslt == 1)
         {
            sprintf(statusmsg,"1-Wire setup is success");
            *Status = STATUS_STEP_COMPLETE;
         }
         else if (rslt == 2)
         {
            sprintf(statusmsg,"1-Wire shorted");
            // check for error last time 
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         else
         {
            sprintf(statusmsg,"1-Wire Net has a hardware problem, abort");
            *Status = STATUS_ERROR_HALT;
         }
         break;

      // release the session on the 1-Wire Net
      case ST_REL_SESSION:
         TMEndSession(session_handle);
         sprintf(statusmsg,"1-Wire Net released");
         *Status = STATUS_STEP_COMPLETE;
         break;      

      // the operation is complete      
      case ST_FINISH:
         sprintf(statusmsg,"Operation complete");
         *Status = STATUS_COMPLETE;
         break;      

      // Select and verify device
      case ST_VERIFY:
         // get the rom from the mib and set
         if (DeviceSetVerify())
         {
            // print rom and then the state
            ReadFromMIB(DD_RomData, tbuf, 8);
            scnt = ROMToStr(&statusmsg[0],tbuf);

            sprintf(&statusmsg[scnt]," Device found and selected");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         { 
            // print rom and then the state
            ReadFromMIB(DD_RomData, tbuf, 8);
            scnt = ROMToStr(&statusmsg[0],tbuf);

            sprintf(&statusmsg[scnt]," Error, could not find device");
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;      

      //Recall EE calibration
      case ST_EE_RECALL:
         // access the device 
         if (TMAccess(session_handle,state_buffer) == 1)
         {
            // send EE recall
            TMTouchByte(session_handle,0xB8);
            sprintf(statusmsg,"Recall EE send successfully");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         { 
            sprintf(statusmsg,"Error, could not access the device");
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;      

      // Start Temperation Conversion
      case ST_TEMP_CONV_1820:
         // access the device 
         if (TMAccess(session_handle,state_buffer) == 1)
         {
            // prepare the strong pullup after next TMTouchByte 
            TMOneWireLevel(session_handle,LEVEL_SET,LEVEL_STRONG_PULLUP,PRIMED_BYTE);

            // send start conversion command 
            TMTouchByte(session_handle,0x44);
            sprintf(statusmsg,"Sent start temperature conversion");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         { 
            sprintf(statusmsg,"Error, could not access the device");
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;      

      // Delay 100ms
      case ST_DELAY_100MS:
         // verify session is still valid
         if (TMValidSession(session_handle) == 1)
         {
            Sleep(100);
            sprintf(statusmsg,"Session valid, delay 100ms complete");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         { 
            sprintf(statusmsg,"Error, session not valid.");
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_HALT;
         }
         break;      
            
      // Read the temperature
      case ST_READ_TEMP_1820:
         // disable the strong pull-up
         TMOneWireLevel(session_handle,LEVEL_SET,LEVEL_NORMAL,PRIMED_NONE);
         if (Read1820Temp(&tmp))
         {
            // set to return value 
            pfloat = prslt;
            *pfloat = tmp;

            sprintf(statusmsg,"Temperature read success: "
                              "%5.1f C, %5.1f F\n",tmp,
                               tmp * 9 / 5 + 32);
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         { 
            sprintf(statusmsg,"Error, could not read temperature");
            *Status = STATUS_ERROR_HALT;
         }
         break;      

      // Set the branch state
      case ST_SET_BRANCH_2406:
         if (Set2406State())
         {
            sprintf(statusmsg,"Branch state set");
            *Status = STATUS_STEP_COMPLETE;
            pint = prslt;
            *pint = TRUE;
         }
         else
         { 
            sprintf(statusmsg,"Error, setting branch state");
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
            pint = prslt;
            *pint = FALSE;
         }
         break;

      // Set the branch state
      case ST_SET_BRANCH_2409:
         if (Set2409State())
         {
            sprintf(statusmsg,"Branch state set");
            *Status = STATUS_STEP_COMPLETE;
            pint = prslt;
            *pint = TRUE;
         }
         else
         { 
            sprintf(statusmsg,"Error, setting branch state");
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
            pint = prslt;
            *pint = FALSE;
         }
         break;      

      // verify the state of a rom switches
      case ST_VERIFY_SWITCH:
         // read the MIB for the array or ROM switches
         obln = ReadFromMIB(DD_InitState, &tbuf[0], sizeof(tbuf));
         if (obln <= 0)
         { 
            sprintf(statusmsg,"Error, could not read init buffer in MIB");
            *Status = STATUS_ERROR_HALT;
         }
         
         // use SubStep to see which one to look at
         if (*SubStep >= (obln / 8))
         {
            sprintf(statusmsg,"Last ROM switch checked");
            *Status = STATUS_STEP_COMPLETE;
         }
         else 
         {
            // print rom and then the state
            scnt = ROMToStr(&statusmsg[0],&tbuf[*SubStep * 8]);

            state = ROMSwitchState(&tbuf[*SubStep * 8]);
            if (state)
               sprintf(&statusmsg[scnt]," switch ON");
            else
               sprintf(&statusmsg[scnt]," switch OFF");
                  
            // collect array of bits for state return
            if (*SubStep < (rslt_max * 8))
               bitacc(WRITE_FUNCTION, state, *SubStep, (uchar *)prslt);   
               
            // increment to look at the next
            *SubStep = *SubStep + 1;
            *Status = STATUS_INPROGRESS;
         }
         break;

      // read the counter specified by the MIB
      case ST_READ_COUNT_2423:
         if (Read2423Counter(&tulong))
         {
            sprintf(statusmsg,"Counter read: %d",tulong);
            *Status = STATUS_STEP_COMPLETE;
            pulong = prslt;
            // check if first reading
            if (pulong[0] == 0)
            {
               pulong[0] = GetTickCount();
               pulong[1] = tulong;
            }
            // second and final reading
            else
            {
               pulong[2] = GetTickCount();
               pulong[3] = tulong;
            }
         }
         else
         { 
            sprintf(statusmsg,"Error, reading counter value");
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;      

      // find the next device and attempt to get a TAG
      case ST_READ_TAG_NEXT:
         *Status = ReadTagNext(SubStep,prslt,rslt_max);
         if (*Status == STATUS_STEP_COMPLETE)
         {
            // check prslt to see if tag data is in it
            puchar = prslt;
            if ((puchar[0] != 0) && (puchar[1] != 0))
            {
               scnt = sprintf(statusmsg,"Tag file found on ");
               // print rom
               trom[0] = 0;
               TMRom(session_handle,state_buffer,trom);
               for (i = 7; i >= 0; i--)
                  scnt += sprintf(statusmsg + scnt,"%02X",trom[i]);
               scnt += sprintf(statusmsg + scnt,"\n");
               output_status(LV_ALWAYS,statusmsg);
            }
            scnt = sprintf(statusmsg,"Tag search complete ");
         }
         else if (*Status == STATUS_ERROR_TRANSIENT)
         {
            scnt = sprintf(statusmsg,"Error reading Tag");
            *ErrorCount = *ErrorCount + 1;
         }
         else if (*Status == STATUS_INPROGRESS)
            sprintf(statusmsg,"Interum step in tag search complete");
         break;

      // read the level specified by the MIB
      case ST_READ_LEVEL_2406:
         if (Read2406Info(TRUE,tbuf))
         {
            sprintf(statusmsg,"Switch State read: %02X",tbuf[0]);
            *Status = STATUS_STEP_COMPLETE;
            if (rslt_max >= 2)
            {
               // calculate the state and add the mask
               puchar = prslt;
               // state in byte 1
               puchar[0] = (uchar)((tbuf[0] >> 2) & 0x03);
               // get mask for byte 2 read the mib entry
               ReadFromMIB(DD_ChannelMask, &puchar[1], 1);
            }
            else
            {   
               *Status = STATUS_ERROR_HALT;
               sprintf(statusmsg,"Result length of %d too small.\n",rslt_max);
            }
         }
         else
         { 
            sprintf(statusmsg,"Error, reading switch info");
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;      

      // read the level specified by the MIB
      case ST_READ_LEVEL_2409:
         if (Read2409Info(tbuf))
         {
            sprintf(statusmsg,"Switch State read: %02X",tbuf[0]);
            *Status = STATUS_STEP_COMPLETE;
            if (rslt_max >= 2)
            {
               // calculate the state and add the mask
               puchar = prslt;
               // state in byte 1
               puchar[0] = (uchar)(((tbuf[0] >> 1) & 0x01) |
                                   ((tbuf[0] >> 2) & 0x02));
               // get mask for byte 2 read the mib entry
               ReadFromMIB(DD_ChannelMask, &puchar[1], 1);
            }
            else
            {   
               *Status = STATUS_ERROR_HALT;
               sprintf(statusmsg,"Result length of %d too small.\n",rslt_max);
            }
         }
         else
         { 
            sprintf(statusmsg,"Error, reading switch info");
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;      

      // read the activity specified by the MIB
      case ST_READ_ACTIVITY_2406:
         if (Read2406Info(FALSE,tbuf))
         {
            sprintf(statusmsg,"Raw switch state read: %02X",tbuf[0]);
            *Status = STATUS_STEP_COMPLETE;
            if (rslt_max >= 2)
            {
               // calculate the state and add the mask
               puchar = prslt;
               // activity state in byte 1
               puchar[0] = (uchar)((tbuf[0] >> 4) & 0x03);
               // get mask for byte 2 read the mib entry
               ReadFromMIB(DD_ChannelMask, &puchar[1], 1);
            }
            else
            {   
               *Status = STATUS_ERROR_HALT;
               sprintf(statusmsg,"Result length of %d too small.\n",rslt_max);
            }
            // clear the activity
            Read2406Info(TRUE,tbuf);
         }
         else
         { 
            sprintf(statusmsg,"Error, reading switch info");
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;

      // read the activity specified by the MIB
      case ST_READ_ACTIVITY_2409:
         if (Read2409Info(tbuf))
         {
            sprintf(statusmsg,"Raw switch state read: %02X",tbuf[0]);
            *Status = STATUS_STEP_COMPLETE;
            if (rslt_max >= 2)
            {
               // calculate the state and add the mask
               puchar = prslt;
               // activity state in byte 1
               puchar[0] = (uchar)((tbuf[0] >> 4) & 0x03);
               // get mask for byte 2 read the mib entry
               ReadFromMIB(DD_ChannelMask, &puchar[1], 1);
            }
            else
            {   
               *Status = STATUS_ERROR_HALT;
               sprintf(statusmsg,"Result length of %d too small.\n",rslt_max);
            }
            // clear the activity (send all-lines-off)
            Send2409Command(0x66,1);
         }
         else
         { 
            sprintf(statusmsg,"Error, reading switch info");
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;      

      default:
         *Status = STATUS_ERROR_HALT;
         sprintf(statusmsg,"1-Wire Step not known\n");
         TMEndSession(session_handle);
         break;
   }

   return *Status;
}

//--------------------------------------------------------------------------
// 1-Wire Utility functions
//--------------------------------------------------------------------------

//--------------------------------------------------------------------------
// Get the rom number from the MIB, set in the driver, verify present.
// Use DD_RomData from MIB.
//
// Return: TRUE - device address set and verified present
//         FALSE - no address available or device not present
//
int DeviceSetVerify(void)
{
   uchar buf[8];
   short ROM[8],i;

   // read the mib entry
   if (ReadFromMIB(DD_RomData, buf, 8) != 8)
      return FALSE;

   // convert to array of shorts         
   for (i = 0; i < 8; i++)
      ROM[i] = buf[i];

   // set the ROM number
   TMRom(session_handle,state_buffer,ROM);  

   return (TMStrongAccess(session_handle,state_buffer) == 1);
}

//----------------------------------------------------------------------
// Read the switch information byte.  Optionally clear the activity
// latch if 'ClearActivity' is TRUE.
//
// Return: TRUE - DS2406 information read and returned in *Info
//         FALSE - error reading DS2406 information
//
//                                                           
int Read2406Info(int ClearActivity, uchar *Info)
{
   short i,rslt;
   uchar send_block[30];
   short send_cnt;

   // access the device
   if (TMAccess(session_handle,state_buffer) == 1)
   {
      // create a block to send the sets the state
      send_cnt = 0; 
      CRC16 = 0;

      // channel access command 
      send_block[send_cnt++] = 0xF5;
      docrc16(0xF5);

      // control bytes
      if (ClearActivity)
      {
         send_block[send_cnt++] = 0xD5;
         docrc16(0xD5);
      }
      else
      {
         send_block[send_cnt++] = 0x55;
         docrc16(0x55);
      }
      send_block[send_cnt++] = 0xFF;
      docrc16(0xFF);

      // read the info byte, dummy and CRC16
      for (i = 0; i < 4; i++)
         send_block[send_cnt++] = 0xFF;

      // now send the block
      rslt = TMBlockStream(session_handle,send_block,send_cnt);
      if (rslt == send_cnt)
      {
         // calculate the CRC16 on the CRC16 bytes to get B001
         for (i = (short)(send_cnt - 4); i < send_cnt; i++)
            docrc16(send_block[i]);

         // get the info value
         *Info = send_block[send_cnt - 4];

         // check result 
         if (CRC16 == 0xB001) 
            return TRUE;
      }
   }

   return FALSE;
}

//----------------------------------------------------------------------
// Read the switch information byte.
//
// Return: TRUE - DS2409 information read and returned in *Info
//         FALSE - error reading DS2409 information
//
//                                                           
int Read2409Info(uchar *Info)
{
   short rslt;
   uchar send_block[30];
   short send_cnt;

   // access the device 
   if (TMAccess(session_handle,state_buffer) == 1)
   {
      // create a block to send the sets the state
      send_cnt = 0; 
      // read status command 
      send_block[send_cnt++] = 0x5A;
      // control byte
      send_block[send_cnt++] = 0x00;
      // status
      send_block[send_cnt++] = 0xFF;
      // confimation
      send_block[send_cnt++] = 0xFF;

      // now send the block
      rslt = TMBlockStream(session_handle,send_block,send_cnt);
      if (rslt == send_cnt)
      {
         // verify status == confirmation byte
         if (send_block[send_cnt - 1] == send_block[send_cnt - 2]) 
         {
            // record the status and return TRUE
            *Info = send_block[send_cnt - 1];
            return TRUE;
         }
      }
   }

   return FALSE;
}

//--------------------------------------------------------------------------
// Attempt to read the temperature of a regular DS1820 and perform
// the high resolution calculation.  
//
// Returns: TRUE, success, value returned in *temperature
//          FALSE, failure to read temperature
//
int Read1820Temp(float *temperature)
{
   uchar send_block[30];
   int tsht, i;
   float tmp,cr,cpc;
   short send_cnt=0;

   // access the device 
   if (TMAccess(session_handle,state_buffer) == 1)
   {
      // create a block to send that reads the temperature
      // read scratchpad command
      send_block[send_cnt++] = 0xBE;
      // now add the read bytes for data bytes and crc8
      for (i = 0; i < 9; i++)
         send_block[send_cnt++] = 0xFF;

      // now send the block
      if (TMBlockStream(session_handle,send_block,send_cnt) == send_cnt)
      {
         // perform the CRC8 on the last 8 bytes of packet
         DOWCRC = 0;
         for (i = send_cnt - 9; i < send_cnt; i++)
            dowcrc(send_block[i]);

         // verify CRC8 is correct
         if (DOWCRC == 0x00)
         {
            // calculate the high-res temperature
            tsht = send_block[1];
            if (send_block[2] & 0x01)
               tsht |= -256;
            tmp = (float)(tsht/2);
            cr = send_block[7];
            cpc = send_block[8];
            if (cpc == 0)
               return FALSE;   
            else
               tmp = tmp - (float)0.25 + (cpc - cr)/cpc;

            *temperature = tmp;
            // success
            return TRUE;
         }
      }
   }

   return FALSE;
}

//--------------------------------------------------------------------------
// Set the latch state of the DS2406.  
// Use DD_ChannelMask, DD_ChannelState from MIB.
//
// Returns: TRUE, success latch state set 
//          FALSE, failure to set latch state
//
int Set2406State(void)
{
   uchar mask, state, info;
   uchar send_block[30];
   short send_cnt=0;

   // read the mib entry for mask and state
   if (ReadFromMIB(DD_ChannelMask, &mask, 1) != 1)
      return FALSE;
   if (ReadFromMIB(DD_ChannelState, &state, 1) != 1)
      return FALSE;

   // check for 0 mask
   if (mask == 0)
      return TRUE;

   // read the current state to effectivily use the mask
   if (!Read2406Info(FALSE,&info))
      return FALSE;

   // create state byte using mask and current state
   info = (uchar)(((~mask & ~info) & 0x03) | state);

   // access the device 
   if (TMAccess(session_handle,state_buffer) == 1)
   {
      CRC16 = 0;
      // create a block to send that reads the counter
      // write status command
      send_block[send_cnt++] = 0x55;
      docrc16(0x55);
      // address of switch state
      send_block[send_cnt++] = 0x07;
      docrc16(0x07);
      send_block[send_cnt++] = 0x00;
      docrc16(0x00);
      // write state
      send_block[send_cnt++] = (uchar)(((~info & 0x03) << 5) | 0x1F);
      docrc16(send_block[send_cnt - 1]);
      // read CRC16
      send_block[send_cnt++] = 0xFF;
      send_block[send_cnt++] = 0xFF;

      // now send the block
      if (TMBlockStream(session_handle,send_block,send_cnt) == send_cnt)
      {
         // perform the CRC16 on the last 2 bytes of packet
         docrc16(send_block[send_cnt-2]);
         docrc16(send_block[send_cnt-1]);

         // verify CRC16 is correct
         if (CRC16 == 0xB001)
            // success
            return TRUE;
      }
   }

   return FALSE;
}


//--------------------------------------------------------------------------
// Set the latch state of the DS2409.  
// Use DD_ChannelMask, DD_ChannelState from MIB.
//
// Returns: TRUE, success latch state set 
//          FALSE, failure to set latch state
//
int Set2409State(void)
{
   uchar mask, state;
   int cmd,xtra;

   // read the mib entry for mask and state
   if (ReadFromMIB(DD_ChannelMask, &mask, 1) != 1)
      return FALSE;
   if (ReadFromMIB(DD_ChannelState, &state, 1) != 1)
      return FALSE;

   // if no channels selected then done
   if ((mask & 0x03) == 0x00)
      return TRUE;

   // decide on command and number of bytes needed
   switch (mask & state & 0x03)
   {
      case 0x00:
         cmd = 0x66; // ALL_LINES_OFF
         xtra = 1;
         break;
      case 0x02:
         cmd = 0x33; // smart ON_AUX;
         xtra = 3;
         break;
      case 0x01:  case 0x03:
         cmd = 0xA5; // DIRECT_ON_MAIN;
         xtra = 1;
         break;
   }

   // send the command and return the result
   return Send2409Command(cmd,xtra);
}

//----------------------------------------------------------------------
// Send and comfirm a command sent to a DS2409.  
//
// Returns: TRUE, success, command sent and verified 
//          FALSE, failure to verify command
//
int Send2409Command(int Cmd, int NumExtra)
{
   uchar send_block[50];
   short send_cnt,i,rslt;

   // access the device 
   if (TMAccess(session_handle,state_buffer) == 1)
   {
      send_cnt = 0; 
      // the command
      send_block[send_cnt++] = (uchar)Cmd;

      // extra bytes and confirmation
      for (i = 0; i < NumExtra; i++)
         send_block[send_cnt++] = 0xFF;

      // send the command string 
      rslt = TMBlockStream(session_handle,send_block,send_cnt);
      if (rslt == send_cnt)
      {
        if (send_block[send_cnt - 1] == Cmd) 
            return TRUE;
      }
   }

   return FALSE;       
}

//--------------------------------------------------------------------------
// Read the counter value from a DS2423.  
// Use DD_InitState from MIB for counter page.
//
// Returns: TRUE, success, read and placed in *Count 
//          FALSE, failure to read counter
//
int Read2423Counter(ulong *Count)
{
   int rt=FALSE;
   uchar send_block[30],CounterPage;
   int address, i;
   short send_cnt=0;
   CRC16 = 0;

   // read the mib entry for counter page
   if (ReadFromMIB(DD_InitState, &CounterPage, 1) != 1)
      return FALSE;

   // access the device 
   if (TMAccess(session_handle,state_buffer) == 1)
   {
      // create a block to send that reads the counter
      // read memory and counter command
      send_block[send_cnt++] = 0xA5;
      docrc16(0xA5);
      // address of last data byte before counter
      address = (CounterPage << 5) + 31;  
      send_block[send_cnt++] = (uchar)(address & 0xFF);
      docrc16((ushort)(address & 0xFF));
      send_block[send_cnt++] = (uchar)(address >> 8);
      docrc16((ushort)(address >> 8));
      // now add the read bytes for data byte,counter,zero bits, crc16
      for (i = 0; i < 11; i++)
         send_block[send_cnt++] = 0xFF;

      // now send the block
      if (TMBlockStream(session_handle,send_block,send_cnt) == send_cnt)
      {
         // perform the CRC16 on the last 11 bytes of packet
         for (i = send_cnt - 11; i < send_cnt; i++)
            docrc16(send_block[i]);

         // verify CRC16 is correct
         if (CRC16 == 0xB001)
         {
            // success
            rt = TRUE;
            // extract the counter value
            *Count = 0;
            for (i = send_cnt - 7; i >= send_cnt - 10; i--)
            {
               *Count <<= 8;
               *Count |= send_block[i];
            }  
         }
      }
   }
   
   // return the result flag rt
   return rt;
}

//--------------------------------------------------------------------------
// Find a device and search for a 'TAGD.000' file.  If the file is present
// then read it and place the contents in 'tag_buf' with a max length of
// 'tag_max'.  Use 'DeviceCheckedForTag' function to not repeat file search
// on known devices.
//
// Returns: STATUS_STEP_COMPLETE - search complete (file may have been found) 
//          STATUS_INPROGRESS - not complete, call again
//
int ReadTagNext(int *SubStep, uchar *tag_buf, int tag_max)
{
   uchar *retry;
   FileEntry tagf;
   short hndl,len,rslt;

   // check the sub step to see if we need to do a first
   if (*SubStep == 0)
   {
      rslt = TMFirst(session_handle,state_buffer);
      *SubStep = 1;
   }
   else
      rslt = TMNext(session_handle,state_buffer);

   // if device found 
   if (rslt >= 1)
   {
      // check if we have already check this device for TAG file
      if (!DeviceCheckedForTag(&retry))
      {
         // attempt to read the TAGD.000 file
         sprintf((char *)&tagf.name[0],"TAGD");
         tagf.extension = 0;
         hndl = TMOpenFile(session_handle,state_buffer,&tagf);
         len = TMReadFile(session_handle,state_buffer,hndl,tag_buf,(short)tag_max);
         TMCloseFile(session_handle,state_buffer,hndl);
         // check results
         if ((hndl >= 0) && (len > 3))
         {
            // success, max retry and return complete
            *retry = RETRY_MAX + 1;
            return STATUS_STEP_COMPLETE;
         }
         else
         {         
            // failure, increase retry and return inprogress
            *retry = (uchar)(*retry + 1);
            // clear any junk in tag_buf
            tag_buf[0] = 0; 
            tag_buf[1] = 0; 
            return STATUS_INPROGRESS;
         }
      }
      // already tried this device or not file compatible 
      else
         return STATUS_INPROGRESS;
   }
   // no parts done so complete
   else
      return STATUS_STEP_COMPLETE;
}


//--------------------------------------------------------------------------
// Check if the current device has been check for a 'tag' file.  Return a 
// pointer to the 'retry' number.  This number is used to re-read a
// device several times to before giving up on it having a 'tag' file.
//
// Returns: TRUE, device has already been check for a tag 
//          FALSE, device not yet checked or not exceed max retries 
//
int DeviceCheckedForTag(uchar **retry_ptr)
{
   short trom[8],s,r;

   // set retry_ptr to null
   *retry_ptr = NULL;

   // read the rom to check
   trom[0] = 0;
   TMRom(session_handle,state_buffer,trom);  
   
   // loop through all of the rom's known
   for (s = 0; s < stack_num; s++)
   {
      for (r = 0; r < 8; r++)
         if (rom_stack[s][r] != trom[r])
            break;
      
      // check for a match
      if (r == 8)
      {
         // set pointer to retries in case need
         *retry_ptr = &retry_stack[s];
         // all reties done? 
         return (retry_stack[s] >= RETRY_MAX);
      }
   }

   // not in stack so add it 
   if (stack_num < ROM_STACK_MAX)
   {
      // insert into the stack
      rom_stack[stack_num][0] = 0;
      TMRom(session_handle,state_buffer,&rom_stack[stack_num][0]);  
      // clear the retry
      retry_stack[stack_num] = 0;
      // set pointer to retries in case need
      *retry_ptr = &retry_stack[stack_num];
      // increase size of stack
      stack_num++;
      // could check if this type is even compatible, return TRUE
   }            

   // new or now room to add to stack
   return FALSE;
}

//--------------------------------------------------------------------------
// Set the current device address to the value passed.  Verify if 
// the specified device is currently connected to the 1-Wire Net.
//
// Returns: TRUE, device is present on the 1-Wire Net 
//          FALSE, device NOT present on the 1-Wire Net 
//
int ROMSwitchState(uchar *rom)
{
   short ROM[8],i;

   // convert to array of shorts         
   for (i = 0; i < 8; i++)
      ROM[i] = rom[i];

   // set the ROM number
   TMRom(session_handle,state_buffer,ROM);  

   return TMStrongAccess(session_handle,state_buffer);
}

//--------------------------------------------------------------------------
// General Utility functions
//--------------------------------------------------------------------------

//--------------------------------------------------------------------------
// Convert the 8 byte arrary to a string representation of a Dallas 1-Wire
// address.  Print least significant byte first.  
//
// Returns: Number of characters added to string not including zero 
//          termination
//
int ROMToStr(char *str, uchar *rom)
{
   int i,scnt=0;
   
   for (i = 7; i >= 0; i--)
      scnt += sprintf(&str[scnt],"%02X",rom[i]);

   return scnt;
}

//--------------------------------------------------------------------------
// Update the Dallas Semiconductor One Wire CRC (DOWCRC) from the global
// variable DOWCRC and the argument.
//
// 'x' - data byte to calculate the 8 bit crc from
//
// Returns: the updated DOWCRC.
//
uchar dscrc_table[] = {
        0, 94,188,226, 97, 63,221,131,194,156,126, 32,163,253, 31, 65,
      157,195, 33,127,252,162, 64, 30, 95,  1,227,189, 62, 96,130,220,
       35,125,159,193, 66, 28,254,160,225,191, 93,  3,128,222, 60, 98,
      190,224,  2, 92,223,129, 99, 61,124, 34,192,158, 29, 67,161,255,
       70, 24,250,164, 39,121,155,197,132,218, 56,102,229,187, 89,  7,
      219,133,103, 57,186,228,  6, 88, 25, 71,165,251,120, 38,196,154,
      101, 59,217,135,  4, 90,184,230,167,249, 27, 69,198,152,122, 36,
      248,166, 68, 26,153,199, 37,123, 58,100,134,216, 91,  5,231,185,
      140,210, 48,110,237,179, 81, 15, 78, 16,242,172, 47,113,147,205,
       17, 79,173,243,112, 46,204,146,211,141,111, 49,178,236, 14, 80,
      175,241, 19, 77,206,144,114, 44,109, 51,209,143, 12, 82,176,238,
       50,108,142,208, 83, 13,239,177,240,174, 76, 18,145,207, 45,115,
      202,148,118, 40,171,245, 23, 73,  8, 86,180,234,105, 55,213,139,
       87,  9,235,181, 54,104,138,212,149,203, 41,119,244,170, 72, 22,
      233,183, 85, 11,136,214, 52,106, 43,117,151,201, 74, 20,246,168,
      116, 42,200,150, 21, 75,169,247,182,232, 10, 84,215,137,107, 53};

uchar dowcrc(uchar x)
{
   DOWCRC = dscrc_table[DOWCRC ^ x];
   return DOWCRC;
}

//--------------------------------------------------------------------------
// Calculate a new CRC16 from the input data shorteger.  Return the current
// CRC16 and also update the global variable CRC16.
//
static short oddparity[16] = { 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0 };

ushort docrc16(ushort data)
{
   data = (ushort)((data ^ (CRC16 & 0xff)) & 0xff);
   CRC16 >>= 8;

   if (oddparity[data & 0xf] ^ oddparity[data >> 4])
     CRC16 ^= 0xc001;

   data <<= 6;
   CRC16   ^= data;
   data <<= 1;
   CRC16   ^= data;

   return CRC16;
}

//--------------------------------------------------------------------------
// Bit utility to read and write a bit in the buffer 'buf'.
//
// 'op'    - operation (1) to set and (0) to read
// 'state' - set (1) or clear (0) if operation is write (1)
// 'loc'   - bit number location to read or write
// 'buf'   - pointer to array of bytes that contains the bit
//           to read or write
//
// Returns: 1   if operation is set (1)
//          0/1 state of bit number 'loc' if operation is reading
//
int bitacc(int op, int state, int loc, uchar *buf)
{
   int nbyt,nbit;

   nbyt = (loc / 8);
   nbit = loc - (nbyt * 8);

   if (op == WRITE_FUNCTION)
   {
      if (state)
         buf[nbyt] |= (uchar)((0x01 << nbit));
      else
         buf[nbyt] &= (uchar)(~(0x01 << nbit));

      return 1;
   }
   else
      return ((buf[nbyt] >> nbit) & 0x01);
}

